'''
Module holding PreProcessor class
'''
from contextlib import contextmanager

import torch
import pandas as pnd

from torch          import tensor
from vector         import MomentumObject3D as v3d
from dask.dataframe import DataFrame        as DDF

from dmu.logging.log_store import LogStore

log=LogStore.add_logger('ecal_calibration:preprocessor')
# --------------------------
class PreProcessor:
    '''
    Class used to process input data into features and target columns.
    The features needed are:

    - row : Row of brem photon in ECAL
    - col : Calumn of brem photon in ECAL
    - area: Index of region of ECAL, 0, 1, 2
    - energy: Energy of brem photon
    - npvs  : Number of primary vertices
    - block : Index representing part of the year when data was collected

    The targer will be called "mu" and will represent the correction needed
    '''
    # ---------------------------------
    def __init__(self, ddf : DDF, cfg : dict):
        '''
        ddf: Dask dataframe with raw data to preprocess
        '''
        self._ddf      = ddf
        self._cfg      = cfg
        self._brem_cut = '(L1_brem + L2_brem > 0.5) & (L1_brem + L2_brem < 1.5)'
        self._neg_tol  = -10
        self._ddf_res  : DDF
        self._nentries = None
    # ---------------------------------
    def _apply_selection(self, ddf : DDF) -> DDF:
        log.info(f'Applying: {self._brem_cut}')
        ddf = ddf.query(self._brem_cut)

        if 'selection' not in self._cfg:
            return ddf

        for selection in self._cfg['selection']:
            log.info(f'Applying: {selection}')
            ddf = ddf.query(selection)

        return ddf
    # ---------------------------------
    def _values(self, kind : str) -> tensor:
        ddf = self.get_data()
        if self._nentries is None:
            df  = ddf.compute()
        else:
            df  = ddf.head(self._nentries)

        if   kind == 'features':
            l_col = [ var for var in df.columns if var != 'mu' ]
        elif kind == 'targets':
            l_col = [ var for var in df.columns if var == 'mu' ]
        else:
            raise ValueError(f'Invalid kind of value: {kind}')

        arr_val = df[l_col].to_numpy()

        return torch.tensor(arr_val, dtype=torch.float32)
    # ---------------------------------
    def get_data(self, partition_size : str = '10MB') -> DDF:
        '''
        Returns dask dataframe after preprocessing, it contains.

        - The features in the class description.
        - The target for regression, labeled as 'mu'

        Parameters
        ----------------------
        partition_size : Partition size in MB. Used to split among workers.
        '''
        if hasattr(self, '_ddf_res'):
            return self._ddf_res

        ddf = self._ddf.dropna()
        ddf = ddf.repartition(partition_size=partition_size)
        ddf = self._apply_selection(ddf=ddf)

        # This meta thing MUST be here
        # Or else dask will try to "guess"
        # the structure by plugging in data
        # that will make the code rise exception
        meta = {
            'row': 'f8',
            'col': 'f8',
            'are': 'f8',
            'eng': 'f8',
            'npv': 'i8',
            'blk': 'i8',
            'mu' : 'f8'
        }

        ddf = ddf.apply(
                PreProcessor.build_features,
                meta = meta,
                axis = 1)

        self._ddf_res = ddf

        return ddf
    # ---------------------------------
    @property
    def features(self) -> tensor:
        '''
        Returns pytorch tensor with features
        '''

        return self._values(kind='features')
    # ---------------------------------
    @property
    def targets(self) -> tensor:
        '''
        Returns pytorch tensor with values of regression targets
        '''

        return self._values(kind='targets')
    # ---------------------------------
    @contextmanager
    def max_entries(self, nentries : int):
        '''
        Context manager that will
        set, temporarily, the number of entries, e.g for tests
        '''
        old            = self._nentries
        self._nentries = nentries
        try:
            yield
        finally:
            self._nentries = old
    # ---------------------------------
    @staticmethod
    def _get_normal(row : pnd.Series) -> v3d:
        pvx = row['B_BPVX']
        pvy = row['B_BPVY']
        pvz = row['B_BPVZ']

        svx = row['B_END_VX']
        svy = row['B_END_VY']
        svz = row['B_END_VZ']

        dr  = v3d(x=svx - pvx, y=svy - pvy, z=svz - pvz)

        return dr / dr.mag
    # ---------------------------------
    @staticmethod
    def _get_momentum(row : pnd.Series, name : str) -> v3d:
        pt = row[f'{name}_PT' ]
        et = row[f'{name}_ETA']
        ph = row[f'{name}_PHI']

        return v3d(pt=pt, eta=et, phi=ph)
    # ---------------------------------
    @staticmethod
    def _get_correction(row : pnd.Series, lepton : str) -> float:
        norm = PreProcessor._get_normal(row=row)
        l1_p = PreProcessor._get_momentum(row=row, name='L1')
        l2_p = PreProcessor._get_momentum(row=row, name='L2')
        kp_p = PreProcessor._get_momentum(row=row, name= 'H')

        # Remove the component alongside normal. i.e. vectors lie on plane
        l1_p = l1_p - norm * norm.dot(l1_p)
        l2_p = l2_p - norm * norm.dot(l2_p)
        kp_p = kp_p - norm * norm.dot(kp_p)

        if   lepton == 'L1':
            lep = l1_p
            oth = l2_p + kp_p
        elif lepton == 'L2':
            lep = l2_p
            oth = l1_p + kp_p
        else:
            raise ValueError(f'Invalid lepton: {lepton}')

        a = lep.dot(oth)
        b = lep.mag ** 2

        return - a/b
    # ---------------------------------
    @staticmethod
    def build_features(
            row         : pnd.Series,
            lep         : str  = None,
            skip_target : bool = False) -> pnd.Series:
        '''
        Builds features and optionally target, needed for training.

        Parameters:
        row        : Pandas series with information from tree branches, one entry.
        lep        : By default None, if passed as L1/L2, it will target this lepton to buld features
        skip_target: By default false, if True it will not calculate "mu". Recommended for prediction mode

        Returns series with features and target needed for training or prediction
        '''
        log.debug('Building features')

        if   lep is not None:
            log.debug(f'Picking up user defined lepton {lep}')
        elif row['L1_brem'] == 1 and row['L2_brem'] != 1:
            lep = 'L1'
        elif row['L1_brem'] != 1 and row['L2_brem'] == 1:
            lep = 'L2'
        else:
            log.info(row)
            raise ValueError('One and only one electron must have brem')

        data        = {}
        data['nam'] = lep
        data['row'] = row[f'{lep}_BREMHYPOROW']
        data['col'] = row[f'{lep}_BREMHYPOCOL']
        data['are'] = row[f'{lep}_BREMHYPOAREA']
        data['eng'] = row[f'{lep}_BREMTRACKBASEDENERGY'] / 1000 # Very large numbers seem to break down training
        data['npv'] = row['nPVs']
        data['blk'] = row['block']
        data['evn'] = row['EVENTNUMBER']

        if not skip_target:
            data['mu' ] = 1000 * PreProcessor._get_correction(row=row, lepton=lep)

        row = pnd.Series(data)

        return row
# --------------------------



'''
Script with code needed to test Corrector class
'''
from importlib.resources import files

import tqdm
import pytest
import pandas as pnd

from vector                        import MomentumObject4D as v4d
from dmu.generic                   import utilities        as gut
from dmu.logging.log_store         import LogStore
from ecal_calibration              import utilities        as cut
from ecal_calibration.preprocessor import PreProcessor
from ecal_calibration.corrector    import Corrector

pytestmark = pytest.mark.skip(
    reason='''
    This part of the code did not produce useful results
    will turn off the tests, due to time constraints
    ''')
# -----------------------------------------------------------
@pytest.fixture(scope='session', autouse=True)
def _initialize():
    LogStore.set_level('ecal_calibration:regressor'   , 20)
    LogStore.set_level('ecal_calibration:preprocessor', 20)
    LogStore.set_level('ecal_calibration:corrector'   , 10)
# -----------------------------------------------------------
def _load_data(name : str) -> dict:
    fpath = files('ecal_calibration_data').joinpath(f'tests/regressor/{name}.json')
    fpath = str(fpath)

    data            = gut.load_json(fpath)
    data['L1_brem'] = data['L1_HASBREMADDED']
    data['L2_brem'] = data['L2_HASBREMADDED']

    return data
# -----------------------------------------------------------
def _get_corrector() -> Corrector:
    kind = 'row_col_are_eng'
    cfg  = cut.load_cfg(name='tests/corrector/simple')
    cfg['saving']['out_dir'] = f'regressor/predict_{kind}'

    cal  = Corrector(cfg=cfg)

    return cal
# -----------------------------------------------------------
def test_calibrate_simple(_dask_client):
    '''
    Tests `corrector` from the Corrector class
    '''

    data = _load_data(name='row')
    sr   = pnd.Series(data)
    sr   = PreProcessor.build_features(
        row        =  sr,
        lep        ='L1',
        skip_target=True)
    cor      = _get_corrector()

    for val in [100, 200, 300, 400, 500]:
        electron = v4d(pt=2250 + val, eta=3.0, phi=1, m=0.511)
        electron = cor.run(electron, row=sr)
# -----------------------------------------------------------
def test_calibrate_are(_dask_client):
    '''
    Tests tests correction for different areas
    '''

    data     = _load_data(name='row')
    sr       = pnd.Series(data)
    sr       = PreProcessor.build_features(
            row        =  sr,
            lep        ='L1',
            skip_target=True)
    cor      = _get_corrector()

    for are in [0, 1, 2]:
        sr['are']= are
        electron = v4d(pt=2250, eta=3.0, phi=1, m=0.511)
        electron = cor.run(electron, row=sr)
# -----------------------------------------------------------
def test_calibrate_benchmark(_dask_client):
    '''
    Tests `corrector` from the Corrector class
    '''

    data     = _load_data(name='row')
    sr       = pnd.Series(data)
    sr       = PreProcessor.build_features(
            row        =  sr,
            lep        ='L1',
            skip_target=True)
    cor      = _get_corrector()

    with LogStore.level('ecal_calibration:corrector', 30):
        for val in tqdm.tqdm(range(10_000), ascii=' -'):
            electron = v4d(
                    px=-10_000 + val,
                    py=-10_000 + val,
                    pz= 43253,
                    e=  43437)

            electron = cor.run(electron, row=sr)
# -----------------------------------------------------------
def test_calibrate_from_map(_dask_client):
    '''
    Tests tests correction using maps
    '''
    data     = _load_data(name='row')
    sr       = pnd.Series(data)
    sr       = PreProcessor.build_features(
            row        =  sr,
            lep        ='L1',
            skip_target=False)

    cor      = _get_corrector()

    for are in [0, 1, 2]:
        sr['are']= are
        electron = v4d(pt=2250, eta=3.0, phi=1, m=0.511)
        electron = cor.run(electron, row=sr, from_nn=False)
# -----------------------------------------------------------



'''
Script with code meant to create JSON files with
toy decays, needed for tests
'''
import os
import argparse
from importlib.resources import files

import numpy
import uproot
import pandas as pnd

from vector import MomentumObject3D as v3d
from dmu.logging.log_store import LogStore

log=LogStore.add_logger('ecal_calibration:make_toy_decays')
# ------------------------------------
class Data:
    '''
    data class
    '''
    nentries = 10_000
    ntup_ver = 'v2'
    ana_dir  = os.environ['ANADIR']
    l_branch = [
        'Kp_0_PT_TRUE',
        'ep_0_PT_TRUE',
        'em_0_PT_TRUE',
        'Kp_0_eta_TRUE',
        'ep_0_eta_TRUE',
        'em_0_eta_TRUE',
        'Kp_0_phi_TRUE',
        'ep_0_phi_TRUE',
        'em_0_phi_TRUE',
        ]

    d_name= {
            'Kp_0_PT_TRUE'  :  'H_PT' ,
            'Kp_0_eta_TRUE' :  'H_ETA',
            'Kp_0_phi_TRUE' :  'H_PHI',
            # ----------
            'ep_0_PT_TRUE'  : 'L1_PT' ,
            'ep_0_eta_TRUE' : 'L1_ETA',
            'ep_0_phi_TRUE' : 'L1_PHI',
            # ----------
            'em_0_PT_TRUE'  : 'L2_PT' ,
            'em_0_eta_TRUE' : 'L2_ETA',
            'em_0_phi_TRUE' : 'L2_PHI'}
# ------------------------------------
def _get_df() -> pnd.DataFrame:
    root_path = f'{Data.ana_dir}/Rapidsim/{Data.ntup_ver}/bpkpee/13TeV/bpkpee_tree.root'
    log.info(f'Reading datafrom: {root_path}')
    ifile = uproot.open(root_path)
    tree  = ifile['DecayTree']
    df    = tree.arrays(Data.l_branch, library='pd', entry_stop=Data.nentries)
    df    = df.rename(columns=Data.d_name)

    return df
# ------------------------------------
def _add_b_vtx(row : pnd.Series) -> pnd.Series:
    ep = v3d(pt=row.L1_PT, eta=row.L1_ETA, phi=row.L1_PHI)
    em = v3d(pt=row.L2_PT, eta=row.L2_ETA, phi=row.L2_PHI)
    kp = v3d(pt=row.H_PT , eta=row.H_ETA , phi=row.H_PHI )

    bp = ep + em + kp
    pos= numpy.random.exponential(scale=10) * bp / bp.mag
    sr = pnd.Series(
            {'B_END_VX' : pos.px,
             'B_END_VY' : pos.py,
             'B_END_VZ' : pos.pz,
             'B_BPVX'   : 0,
             'B_BPVY'   : 0,
             'B_BPVZ'   : 0})

    return sr
# ------------------------------------
def _reformat_df(df : pnd.DataFrame) -> pnd.DataFrame:
    df_vtx = df.apply(_add_b_vtx, axis=1)

    df     = df.join(df_vtx)
    df     = _add_lepton_columns(df=df, lepton='L1')
    df     = _add_lepton_columns(df=df, lepton='L2')
    df     = _add_event_columns(df=df)

    return df
# ------------------------------------
def _add_event_columns(df : pnd.DataFrame) -> pnd.DataFrame:
    size        = len(df)
    df['nPVs' ] = numpy.random.poisson(lam=3, size=size)
    df['block'] = numpy.random.randint(1, 9,  size=size)

    return df
# ------------------------------------
def _add_lepton_columns(df : pnd.DataFrame, lepton : str) -> pnd.DataFrame:
    size   = len(df)

    df[f'{lepton}_BREMHYPOAREA'] = numpy.random.choice([0, 1, 2], size=size)
    df[f'{lepton}_brem'        ] = numpy.random.choice([0, 1   ], size=size, p=[0.2, 0.8])
    df[f'{lepton}_BREMHYPOCOL' ] = numpy.random.randint(1, 60, size=size)
    df[f'{lepton}_BREMHYPOROW' ] = numpy.random.randint(1, 60, size=size)
    df[f'{lepton}_BREMTRACKBASEDENERGY' ] = numpy.random.exponential(scale=30_000, size=size)

    return df
# ------------------------------------
def _parse_args():
    parser = argparse.ArgumentParser(description='Script used to produce JSON file with toy decays needed for running tests')
    parser.add_argument('-n', '--nentries', type=int, help='Number of entries to process', default=Data.nentries)
    args = parser.parse_args()

    Data.nentries = args.nentries
# ------------------------------------
def main():
    '''
    Start here
    '''
    _parse_args()

    df = _get_df()
    df = _reformat_df(df=df)

    out_dir = files('ecal_calibration_data').joinpath('tests/data')
    os.makedirs(out_dir, exist_ok=True)

    out_path = f'{out_dir}/toy_decays.parquet'
    log.info(f'Sending output to: {out_path}')
    df.to_parquet(out_path, compression='snappy')
# ------------------------------------
if __name__ == '__main__':
    main()



'''
Module with code needed to retrieve X, Y position from ECAL cell ID
'''

from importlib.resources import files
import pandas as pnd
from dmu.logging.log_store import LogStore

log=LogStore.add_logger('ecal_calibration:calo_translator')
# --------------------------------
def _cast_column(column, ctype) -> pnd.Series:
    column = pnd.to_numeric(column, errors='coerce')
    column = column.fillna(-100_000)
    column = column.astype(ctype)

    return column
# --------------------------------
def get_data() -> pnd.DataFrame:
    '''
    Returns pandas dataframe with x,y and row and column values
    '''
    data_path = files('ecal_calibration_data').joinpath('brem_correction/coordinates.csv')
    df      = pnd.read_csv(data_path)
    df['a'] = _cast_column(df.a, int)
    df['x'] = _cast_column(df.x, float)
    df['y'] = _cast_column(df.y, float)
    df['z'] = _cast_column(df.z, float)
    df['r'] = _cast_column(df.r, int)
    df['c'] = _cast_column(df.c, int)

    return df
# ------------------------------------------------------
def from_id_to_xy(row : int = None, col : int = None, det : str = None, area : int = None) -> tuple[float,float]:
    '''
    Function taking position in ECAL
    Parameters
    --------------
    row: Row index
    col: Column index
    det: Name of subdetector, Inner, Middle, Outer
    area: Index of subdetector 0 (Inner), 1 (Middle), 2 (Outer)

    Returns
    --------------
    tuple[float,float]: With the x,y coordinates, if the region in the detector is fully specified.
    df : Pandas dataframe with the location information, if the region is not fully specified
    '''
    if det not in [
            None,
            'Inner',
            'Middle',
            'Outer']:
        raise ValueError(f'Invalid subdetector name: \"{det}\"')

    df = get_data()
    if area is not None:
        df = df[df.a==area]

    if det is not None:
        df = df[df.n==det]

    if row is not None:
        df = df[df.r==row]

    if col is not None:
        df = df[df.c==col]

    are = area

    has_row = row is not None
    has_col = col is not None
    has_det = det is not None
    has_are = are is not None

    full_description = has_row and has_col and (has_det or has_are)

    if not full_description and len(df) <= 1:
        log.error(df)
        raise ValueError('Expected dataframe with more than one row')

    if not full_description and len(df) > 1:
        return df

    if full_description and len(df) != 1:
        log.error(df)
        log.info(f'{"Row":<10}{row}')
        log.info(f'{"Col":<10}{col}')
        log.info(f'{"Det":<10}{det}')
        log.info(f'{"Reg":<10}{are}')
        raise ValueError('Expected dataframe with one row')

    x = df.iloc[0]['x']
    y = df.iloc[0]['y']

    return x, y
# ------------------------------------------------------



'''
Module containing the Regressor class
'''
# pylint: disable=too-many-instance-attributes

import os

import pandas            as pnd
import matplotlib.pyplot as plt
import mplhep
import torch
import numpy

from torch             import nn
from torch             import optim
from torch             import Tensor
from scipy.interpolate import griddata

from dask.dataframe           import DataFrame as DDF
from dmu.logging.log_store    import LogStore
from ecal_calibration.network import Network, ConstantModel
from ecal_calibration         import calo_translator as ctran

log=LogStore.add_logger('ecal_calibration:regressor')
# ---------------------------------------------
class Regressor:
    '''
    Class used to train a regressor to _learn_ energy
    corrections
    '''
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    # ---------------------------------------------
    def __init__(self, ddf_tr : DDF, ddf_ts : DDF, cfg : dict):
        '''
        Parameters
        -------------------
        ddf_tr : Dask dataframe storing training dataset
        ddf_ts : Dask dataframe storing testing dataset
        cfg    : Dictionary holding configuration
        '''
        plt.style.use(mplhep.style.LHCb2)

        self._ddf_tr = ddf_tr
        self._ddf_ts = ddf_ts
        self._cfg    = cfg
        self._out_dir= Regressor.get_out_dir(cfg=cfg, create=True)

        self._d_area = {0 : 'Outer', 1 : 'Middle', 2 : 'Inner'}
        self._d_color= {0 : 'blue' , 1 : 'green' , 2 : 'red'}
        self._d_var  = {'Real' : 'mu', 'Predicted' : 'mu_pred'}

        self._net : Network
    # ---------------------------------------------
    def _save_regressor(self, regressor : Network) -> None:
        out_path = f'{self._out_dir}/model.pth'
        log.info(f'Saving model to: {out_path}')
        torch.save(regressor, out_path)
    # ---------------------------------------------
    def train(self, constant_target : float = None) -> None:
        '''
        Will train the regressor

        Parameters
        -------------
        constant_target (float) : By default None. If passed, will create network that outputs always this value. Used for debugging
        '''

        features, targets   = Regressor.get_tensors(cfg = self._cfg, ddf = self._ddf_tr)
        nsamples, nfeatures = features.shape

        log.info(f'Training with {nsamples} samples')

        if constant_target is None:
            net = Network(nfeatures=nfeatures, model=self._cfg['model'])
        else:
            net = ConstantModel(target=constant_target)

        net      = Regressor.move_to_gpu(net)
        features = Regressor.move_to_gpu(features)
        targets  = Regressor.move_to_gpu(targets)

        criterion = nn.MSELoss()

        cfg_trn   = self._cfg['train']

        learning_rate = cfg_trn['lr']
        log.debug(f'Using learning_rate: {learning_rate}')
        optimizer = optim.Adam(net.parameters(), lr=learning_rate)
        for epoch in range(cfg_trn['epochs']):
            net.train()
            optimizer.zero_grad()
            outputs = net(features)
            loss    = criterion(outputs, targets)
            loss.backward()
            optimizer.step()

            if epoch % 2000 == 0:
                log.info(f'Epoch {epoch}, Loss: {loss.item():.4f}')

        self._save_regressor(regressor=net)

        net.eval()
        self._net = net
    # ---------------------------------------------
    def _add_xy(self, row : pnd.Series, var : str) -> pnd.Series:
        r = row['row']
        c = row['col']
        a = row['are']

        x, y = ctran.from_id_to_xy(row=int(r), col=int(c), area=int(a))

        return x if var == 'x' else y
    # ---------------------------------------------
    def test(self) -> None:
        '''
        Runs comparison of predicted
        '''
        self._net = Regressor.load(model_dir = self._out_dir)
        if self._net is None:
            log.info('Model not found, training it')
            self.train()

        self._net     = Regressor.move_to_gpu(self._net)
        l_fea         = self._cfg['features']
        ddf           = self._ddf_ts
        ddf           = ddf[ddf['row'] > 1]
        ddf           = ddf[ddf['col'] > 1]

        df            = ddf.compute()
        arr_fea       = df[l_fea].values
        features      = torch.tensor(arr_fea, dtype=torch.float32)

        df['mu_pred'] = self.predict(features=features) / 1000.
        df['mu']      = df['mu'] / 1000.
        df['x']       = df.apply(self._add_xy, args=('x',), axis=1)
        df['y']       = df.apply(self._add_xy, args=('y',), axis=1)

        self._plot_corrections(df=df)
        self._plot_by_energy(df=df)
        self._plot_by_area(df=df)

        self._plot_by_npvs(df=df, kind=     'Real')
        self._plot_by_npvs(df=df, kind='Predicted')

        self._plot_by_block(df=df, kind=     'Real')
        self._plot_by_block(df=df, kind='Predicted')

        self._plot_in_xy(df=df, kind=     'Real')
        self._plot_in_xy(df=df, kind='Predicted')
    # ---------------------------------------------
    def _plot_in_xy(self, df : pnd.DataFrame, kind : str) -> None:
        var = self._d_var[kind]
        grid_x, grid_y = numpy.mgrid[-3000:+3000:100j, -3000:+3000:100j]

        levels  = numpy.linspace(0.0, 1.5, 50)
        grid_z  = griddata((df['x'], df['y']), df[var], (grid_x, grid_y), method='cubic')
        contour = plt.contourf(grid_x, grid_y, grid_z, levels=levels, cmap='viridis')

        plt.colorbar(contour, label=kind)
        plt.xlabel('x')
        plt.ylabel('y')
        plt.title(r'$\mu$ vs $x$ and $y$')
        plt.tight_layout()
        plt.savefig(f'{self._out_dir}/xy_correction_{kind}.png')
        plt.close()
    # ---------------------------------------------
    def _plot_corrections(self, df : pnd.DataFrame) -> None:
        nentries = len(df)

        ax = None
        for area, df_area in df.groupby('are'):
            area = int(area)
            name = self._d_area[area]
            color= self._d_color[area]

            ax   = df_area.plot.scatter(x='mu', y='mu_pred', s=1, label=name, ax=ax, color=color)

        plt.xlabel('Real')
        plt.ylabel('Predicted')
        plt.title(f'Entries: {nentries}')
        plt.grid()
        plt.savefig(f'{self._out_dir}/prediction_vs_real.png')
        plt.close()
    # ---------------------------------------------
    def _plot_by_area(self, df : pnd.DataFrame) -> None:
        for area, df_area in df.groupby('are'):
            area = int(area)
            name = self._d_area[area]

            ax = None
            ax = df_area['mu'     ].plot.hist(bins=20, range=[0.0, 1.5], label='Real'     , alpha   =   0.3, ax=ax, color='blue')
            ax = df_area['mu_pred'].plot.hist(bins=20, range=[0.0, 1.5], label='Predicted', histtype='step', ax=ax, color='red' )

            plt.title(f'Region: {name}')
            plt.xlabel('Correction')
            plt.ylabel('Entries')
            plt.legend()
            plt.savefig(f'{self._out_dir}/area_{name}.png')
            plt.close()
    # ---------------------------------------------
    def _plot_by_energy(self, df : pnd.DataFrame) -> None:
        ax = None
        ax = df.plot.scatter('eng', 'mu'     , label='Real'     , color='blue', s=1, ax=ax)
        ax = df.plot.scatter('eng', 'mu_pred', label='Predicted', color='red' , s=1, ax=ax)

        plt.legend()
        plt.xlabel(r'$E(\gamma)$ MeV')
        plt.ylabel(r'$\mu$')
        plt.savefig(f'{self._out_dir}/corr_vs_energy.png')
        plt.close()
    # ---------------------------------------------
    def _plot_by_npvs(self, df : pnd.DataFrame, kind : str) -> None:
        ax  = None
        var = self._d_var[kind]
        for npv, df_npv in df.groupby('npv'):
            npv = int(npv)

            if npv % 2 == 0:
                continue

            if npv > 10:
                break

            ax = df_npv[var].plot.hist(label=f'nPV={npv}', bins=10, range=[0.0, 1.5], histtype='step', density=True, ax=ax)

        plt.legend()
        plt.title(kind)
        plt.xlabel(r'Correction')
        plt.ylabel(r'$\mu$')

        plt.savefig(f'{self._out_dir}/corr_vs_npv_{kind}.png')
        plt.close()
    # ---------------------------------------------
    def _plot_by_block(self, df : pnd.DataFrame, kind : str) -> None:
        ax  = None
        var = self._d_var[kind]
        for blk, df_blk in df.groupby('blk'):
            blk= int(blk)
            ax = df_blk[var].plot.hist(label=f'Block={blk}', bins=10, range=[0.0, 1.5], histtype='step', density=True, ax=ax)

        plt.legend()
        plt.title(kind)
        plt.xlabel(r'Correction')
        plt.ylabel(r'$\mu$')

        plt.savefig(f'{self._out_dir}/corr_vs_block_{kind}.png')
        plt.close()
    # ---------------------------------------------
    def predict(self, features : Tensor) -> numpy.ndarray:
        '''
        Runs prediction of targets and returns them as a numpy array

        If model does not exist it will train it with the data passed in the initializer

        Parameters
        ---------------
        features: Tensor with features to predict from

        Returns
        ---------------
        Numpy array with values of predicted targets
        '''
        self._net = Regressor.load(model_dir = self._out_dir)
        if self._net is None:
            log.info('Model not found, training it')
            self.train()

        self._net = Regressor.move_to_gpu(self._net)
        features  = Regressor.move_to_gpu(features)
        targets   = self._net(features)
        targets   = targets.cpu()

        return targets.detach().numpy()
    # ---------------------------------------------
    @staticmethod
    def move_to_gpu(x):
        '''
        Will move tensor to GPU in order to do training, prediction, etc
        '''
        if not torch.cuda.is_available():
            log.warning('Cannot move object to GPU, GPU not available?')
            return x

        log.debug('Moving object to GPU')

        x = x.to(Regressor.device)

        return x
    # ---------------------------------------------
    @staticmethod
    def get_out_dir(cfg : dict, create : bool = False) -> str:
        '''
        This function assumes that the path will **start** in ANADIR

        Parameters
        -----------------
        cfg   : Config with 'saving:out_dir' entry specifying path
        create: If true, it will try to create the directory
        '''
        ana_dir = os.environ['ANADIR']
        out_dir = cfg['saving']['out_dir']
        out_dir = f'{ana_dir}/{out_dir}'

        log.info(f'Using output directory: {out_dir}')
        if create:
            os.makedirs(out_dir, exist_ok=True)

        return out_dir
    # ---------------------------------------------
    @staticmethod
    def get_tensors(cfg : dict, ddf : DDF) -> tuple[Tensor,Tensor]:
        '''
        Provides features and target tensors in a tuple

        Parameters
        -----------------
        cfg : Dictionary storing configuration
        ddf : Dask dataframe with data
        '''
        target     = cfg['target']
        l_feat     = cfg['features']

        log.debug(f'Using features: {l_feat}')
        log.debug(f'Using target  : {target}')

        df         = ddf.compute()
        arr_target = df[target].to_numpy()
        arr_feat   = df[l_feat].values

        features   = torch.tensor(arr_feat, dtype=torch.float32)
        targets    = torch.tensor(arr_target, dtype=torch.float32)
        targets    = targets.unsqueeze(1)

        log.debug(f'Features shape: {features.shape}')
        log.debug(f'Targets shape: {targets.shape}')

        return features, targets
    # ---------------------------------------------
    @staticmethod
    def load(model_dir : str) -> Network|None:
        '''
        Parameters
        ----------------
        model_dir: Directory where model.pth is found

        Returns
        ----------------
        Network with model
        '''
        model_path = f'{model_dir}/model.pth'
        log.debug(f'Picking model from: {model_path}')

        if not os.path.isfile(model_path):
            log.info(f'Model not found in: {model_path}')
            return None

        net = torch.load(model_path, map_location='cpu', weights_only=False)
        net.eval()

        return net
# ---------------------------------------------



'''
Module holding code meant to be reused elsewhere
'''

from importlib.resources import files

from torch               import Tensor
from dask                import dataframe
from dask.dataframe      import DataFrame as DDF

import dmu.generic.utilities as gut

# ------------------------------------
def load_cfg(name : str) -> dict:
    '''
    Loads YAML file with configuration and returns dictionary.

    Parameters:

    name (str) : String representing part of the path to config file, e.g. regressor/simple for .../regressor/simple.yaml
    '''

    config_path = files('ecal_calibration_data').joinpath(f'{name}.yaml')
    config_path = str(config_path)
    data        = gut.load_json(config_path)

    return data
# ---------------------------------------------
def _inject_bias(ddf : DDF, bias : float, kind : str) -> DDF:
    '''
    This function scales the momentum components of the lepton by the `bias` factor
    This is done only when the electrons have brem associated, i.e. L*_brem == 1

    Parameters
    --------------
    kind (str) : Type of bias
        flat: bias is uncorrelated with anything
        row : Correlation with the row
    '''
    for lep in ['L1', 'L2']:
        if   kind == 'flat':
            ddf[f'{lep}_PT'] = ddf[f'{lep}_PT'] + ddf[f'{lep}_PT'] * ddf[f'{lep}_brem'] * (bias - 1)
        elif kind == 'row' :
            ddf[f'{lep}_PT'] = ddf[f'{lep}_PT'] + ddf[f'{lep}_PT'] * ddf[f'{lep}_brem'] * (bias - ddf[f'{lep}_BREMHYPOROW'] / 60.)
        elif kind == 'row_col' :
            ddf[f'{lep}_PT'] = ddf[f'{lep}_PT'] + ddf[f'{lep}_PT'] * ddf[f'{lep}_brem'] * (bias - ddf[f'{lep}_BREMHYPOROW'] / 60. * ddf[f'{lep}_BREMHYPOCOL']/ 60.)
        elif kind == 'row_col_are' :
            ddf[f'{lep}_PT'] = ddf[f'{lep}_PT'] + ddf[f'{lep}_PT'] * ddf[f'{lep}_brem'] * (bias - ddf[f'{lep}_BREMHYPOROW'] / 60. * ddf[f'{lep}_BREMHYPOCOL']/ 60.) * ddf[f'{lep}_BREMHYPOAREA']
        elif kind == 'row_col_are_eng' :
            ddf[f'{lep}_PT'] = ddf[f'{lep}_PT'] + ddf[f'{lep}_PT'] * ddf[f'{lep}_brem'] * (bias - ddf[f'{lep}_BREMHYPOROW'] / 60. * ddf[f'{lep}_BREMHYPOCOL']/ 60.) * ddf[f'{lep}_BREMHYPOAREA'] * ddf[f'{lep}_BREMTRACKBASEDENERGY'] / 10_000.
        else:
            raise ValueError(f'Invalid bias: {kind}')

    return ddf
# ------------------------------------
def normalize_tensor(x : Tensor) -> Tensor:
    '''
    Makes sure the mean of the distribution is zero and the standard deviation is 1
    '''
    mean = x.mean()
    std  = x.std()
    x    = (x - mean) / std

    return x
# ------------------------------------
def get_ddf(
        name : str,
        bias : float|None,
        kind : str) -> DDF:
    '''
    Returns Dask DataFrame with toy data, used for tests

    Parameters
    ---------------
    name (str)  : Name of file, e.g. real_data/fake_data
    bias (float): Numerical value of bias, if flat, should be around 1, 1 will be no bias. Pass None, if no bias is will be used
    kind (str)  : Type of bias, `flat` for same bias for all electrons, `row` for row dependent one.
    '''
    data_path = files('ecal_calibration_data').joinpath(f'tests/data/{name}.parquet')
    data_path = str(data_path)
    ddf       = dataframe.read_parquet(data_path)

    if bias is None:
        return ddf

    ddf = _inject_bias(ddf=ddf, bias=bias, kind=kind)

    return ddf
# ------------------------------------



'''
Module used to hold:

Network: Class meant to represent a neural network, used for actual prediction
ConstantModel: Used for debugging purposes, outputs the same value for any input
'''

import torch
from torch import nn
from torch import Tensor

from dmu.logging.log_store import LogStore

log=LogStore.add_logger('ecal_calibration:network')
# --------------------------------------
class Network(nn.Module):
    '''
    Class wrapping pytorch (abstract?) newtwork
    '''
    # ------------------------------
    def __init__(self, nfeatures : int, model : str):
        '''
        nfeatures (int): Number of features, needed to build first layer
        model          : Name of model, e.g. v1
        '''
        self._nfeatures = nfeatures

        super().__init__()

        fun = getattr(self, f'_model_{model}', None)
        if fun is None:
            raise ValueError(f'No model {model} found')

        self.model      = fun()
    # ------------------------------
    def _model_v1(self) -> nn.Sequential:
        log.info('Using model v1')

        model = nn.Sequential(
            nn.Linear(self._nfeatures, 6),
            nn.ReLU(),
            nn.Linear(6,               1)
        )

        return model
    # ------------------------------
    def _model_v2(self) -> nn.Sequential:
        log.info('Using model v2')

        model = nn.Sequential(
            nn.Linear(self._nfeatures, 6),
            nn.ReLU(),
            nn.Linear(6              , 6),
            nn.ReLU(),
            nn.Linear(6              , 1)
        )

        return model
    # ------------------------------
    def _model_v3(self) -> nn.Sequential:
        log.info('Using model v3')

        model = self.model = nn.Sequential(
            nn.Linear(self._nfeatures, 10),
            nn.ReLU(),
            nn.Linear(10             , 10),
            nn.ReLU(),
            nn.Linear(10             ,  1)
        )

        return model
    # ------------------------------
    def _model_v4(self) -> nn.Sequential:
        log.info('Using model v4')

        model = self.model = nn.Sequential(
            nn.Linear(self._nfeatures,  6),
            nn.ReLU(),
            nn.Linear(6              ,  6),
            nn.ReLU(),
            nn.Linear(6              ,  6),
            nn.ReLU(),
            nn.Linear(6              ,  1)
        )

        return model
    # ------------------------------
    def _model_v5(self) -> nn.Sequential:
        log.info('Using model v5')

        model = self.model = nn.Sequential(
            nn.Linear(self._nfeatures, 15),
            nn.ReLU(),
            nn.Linear(15             , 15),
            nn.ReLU(),
            nn.Linear(15             ,  1)
        )

        return model
    # ------------------------------
    def _model_v6(self) -> nn.Sequential:
        log.info('Using model v6')

        model = self.model = nn.Sequential(
            nn.Linear(self._nfeatures, 15),
            nn.ReLU(),
            nn.Linear(15             , 15),
            nn.ReLU(),
            nn.Linear(15             , 15),
            nn.ReLU(),
            nn.Linear(15             ,  1)
        )

        return model
    # ------------------------------
    def _model_v7(self) -> nn.Sequential:
        log.info('Using model v7')

        model = self.model = nn.Sequential(
            nn.Linear(self._nfeatures, 20),
            nn.ReLU(),
            nn.Linear(20             , 20),
            nn.ReLU(),
            nn.Linear(20             , 20),
            nn.ReLU(),
            nn.Linear(20             ,  1)
        )

        return model
    # ------------------------------
    def forward(self, x : Tensor) -> Tensor:
        '''
        Evaluates the features through the model
        '''
        return self.model(x)
# --------------------------------------
class ConstantModel(nn.Module):
    '''
    Model used for debugging purposes, it outputs a given constant
    '''
    # ------------------------------
    def __init__(self, target : float):
        super().__init__()
        self.value = nn.Parameter(torch.tensor(target))
    # ------------------------------
    def forward(self, x: Tensor) -> Tensor:
        '''
        Returns required target, no training needed
        '''
        return self.value.expand(x.shape[0], 1)
# --------------------------------------



'''
Script with code needed to test Calibration class
'''
from importlib.resources import files

import numpy
import pytest
import matplotlib.pyplot as plt

from torch                         import Tensor
from dask.distributed              import Client
from dmu.generic                   import utilities as gut
from ecal_calibration.preprocessor import PreProcessor
from ecal_calibration.regressor    import Regressor
from ecal_calibration              import utilities as cut


pytestmark = pytest.mark.skip(
    reason='''
    This part of the code did not produce useful results
    will turn off the tests, due to time constraints
    ''')
# -----------------------------------------------------------
def _load_data(name : str) -> dict:
    fpath = files('ecal_calibration_data').joinpath(f'tests/regressor/{name}.json')
    fpath = str(fpath)
    data  = gut.load_json(fpath)

    return data
# -----------------------------------------------------------
def _plot_target_vs_features(arr_pred : numpy.ndarray, features : Tensor) -> None:
    arr_feat = features.numpy()
    arr_feat = arr_feat.T[0] # The zeroth feature is the one the target depends on, for now

    plt.scatter(arr_feat, arr_pred, s=1)
    plt.xlabel('Feature')
    plt.ylabel('Prediction')
    plt.grid()
    plt.show()
# -----------------------------------------------------------
def _plot_targets(
        pred : numpy.ndarray,
        real : numpy.ndarray,
        corr : float) -> None:

    minx = min(numpy.min(pred), numpy.min(real))
    maxx = max(numpy.max(pred), numpy.max(real))

    plt.hist(real, bins=80, range=[minx, maxx], label='Training' , alpha   =   0.3)
    plt.hist(pred, bins=80, range=[minx, maxx], label='Predicted', histtype='step')

    if corr is not None:
        plt.axvline(x=corr, label='Correction', ls=':', color='red')

    plt.legend()
    plt.xlabel('Real')
    plt.ylabel('Predicted')
    plt.show()
# -----------------------------------------------------------
def test_flat_bias():
    '''
    Simplest test for calibration with biased data
    '''
    cfg = cut.load_cfg(name='tests/preprocessor/simple')

    ddf = cut.get_ddf(name='fake_data', bias=1.1, kind='flat')
    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()

    cfg = cut.load_cfg(name='tests/regressor/simple')
    obj = Regressor(ddf_tr=ddf, ddf_ts=ddf, cfg=cfg)
    obj.train()
# -----------------------------------------------------------
def test_loader():
    '''
    Tests loading existing model
    '''
    cfg = cut.load_cfg(name='tests/preprocessor/simple')

    ddf = cut.get_ddf(bias=1.1, kind='flat')
    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()

    cfg = cut.load_cfg(name='tests/regressor/simple')
    obj = Regressor(ddf_tr=ddf, ddf_ts=ddf, cfg=cfg)
    obj.load()
# -----------------------------------------------------------
@pytest.mark.parametrize('bias', [0.5, 0.8, 1.0, 1.2, 1.4])
def test_constant_predict(bias : float):
    '''
    Meant to test everything around network by:

    - Introducing data with constant (not dependent on features) bias
    - Training a constant model that outputs that bias
    '''
    corr= 1.0 / bias

    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='fake_data', bias=1.1, kind='flat')
    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()

    cfg = cut.load_cfg(name='tests/regressor/simple')
    cfg['train']['epochs']   = 200
    cfg['saving']['out_dir'] = 'regressor/constant_predict'

    obj = Regressor(ddf_tr=ddf, ddf_ts=ddf, cfg=cfg)
    obj.train(constant_target=corr)

    pred= obj.predict(features=pre.features)

    assert numpy.allclose(pred, corr, rtol=1e-5)
# -----------------------------------------------------------
@pytest.mark.parametrize('bias', [0.5, 0.8, 1.0, 1.2, 1.4])
def test_predict_flat_bias(bias : float):
    '''
    Meant to test whole pipeline by:

    - Introducing data with constant (not dependent on features) bias
    - Training a real model
    '''
    corr= 1.0 / bias

    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='fake_data', bias=1.1, kind='flat')
    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()

    cfg = cut.load_cfg(name='tests/regressor/simple')
    cfg['train']['epochs']   =  2_000
    cfg['input']['nentries'] = 15_000
    cfg['saving']['out_dir'] = 'regressor/predict_flat_bias'

    obj = Regressor(ddf_tr=ddf, ddf_ts=ddf, cfg=cfg)
    obj.train()

    pred= obj.predict(features=pre.features)
    real= pre.targets.numpy()
    _plot_targets(pred=pred, real=real, corr=corr)
# -----------------------------------------------------------
@pytest.mark.parametrize('kind', ['row_col_are_eng'])
def test_predict_bias(_dask_client : Client, kind : str):
    '''
    Meant to test everything by:

    - Training a real model that outputs the correction
    '''
    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='fake_data', bias=None, kind=kind)

    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()
    ddf_tr, ddf_ts = ddf.random_split([0.95, 0.05], random_state=42)

    cfg = cut.load_cfg(name='tests/regressor/simple')
    cfg['train']['epochs']   = 40000
    cfg['saving']['out_dir'] = f'regressor/predict_{kind}'

    obj = Regressor(ddf_tr=ddf_tr, ddf_ts=ddf_ts, cfg=cfg)
    obj.test()
# -----------------------------------------------------------



'''
Module testing PreProcessor class
'''
import os
import logging
import numpy
import torch
import pytest
import pandas            as pnd
import matplotlib.pyplot as plt

from dask.distributed              import Client
from dmu.logging.log_store         import LogStore
from ecal_calibration.preprocessor import PreProcessor
from ecal_calibration              import utilities    as cut

log = LogStore.add_logger('ecal_calibration:test_preprocessor')

pytestmark = pytest.mark.skip(
    reason='''
    This part of the code did not produce useful results
    will turn off the tests, due to time constraints
    ''')
# -----------------------------------------
class Data:
    '''
    Data class
    '''
    out_dir = '/tmp/tests/ecal_calibration/preprocessor'
    columns = {'row', 'col', 'are', 'eng', 'npv', 'blk', 'mu'}
    d_feat  = {'eng' : r'$E(\gamma)$[GeV]', 'row' : 'Row', 'col' : 'Column', 'are' : 'ECAL region', 'npv' : 'nPVs', 'blk' : 'Block'}
# -----------------------------------------
@pytest.fixture(scope='session', autouse=True)
def _initialize():
    os.makedirs(Data.out_dir, exist_ok=True)
    logging.getLogger("PIL.PngImagePlugin").setLevel(logging.WARNING)
    logging.getLogger("matplotlib.font_manager").setLevel(logging.WARNING)
# ---------------------------------------------
def _plot_df(
        df        : pnd.DataFrame,
        test_name : str,
        corr      : float) -> None:

    _plot_features(df=df, test_name=test_name)
    _plot_bias(  df=df, test_name=test_name, corr=corr)
    _plot_var_mu(df=df, test_name=test_name)
# ---------------------------------------------
def _plot_var_mu(df : pnd.DataFrame, test_name : str) -> None:
    out_dir = f'{Data.out_dir}/{test_name}'
    os.makedirs(out_dir, exist_ok=True)
    for column, name in Data.d_feat.items():
        df.plot.scatter(x=column, y='mu', s=1)

        plt.legend()
        plt.grid()
        plt.xlabel(name)
        plt.ylabel(r'$\mu$')
        plt.ylim(0.2, 1.8)
        if name == 'eng':
            plt.xlim(bottom=0.0)
        plt.savefig(f'{out_dir}/{column}_mu.png')
        plt.close()
# ---------------------------------------------
def _plot_bias(df : pnd.DataFrame, test_name : str, corr : float) -> None:
    out_dir = f'{Data.out_dir}/{test_name}'
    os.makedirs(out_dir, exist_ok=True)

    ax = None
    for area, df_area in df.groupby('are'):
        ax = df_area['mu'].plot.hist(
                range=[0,2],
                bins=101,
                label=f'Region {area}',
                histtype='step',
                ax=ax,
                density=False)

    if corr is not None:
        plt.axvline(x=corr, ls=':', label='expected', color='red')

    plt.legend()
    plt.xlabel(r'$\mu$')
    plt.ylabel('Entries')
    plt.grid()
    plt.savefig(f'{out_dir}/mu.png')
    plt.close()
# ---------------------------------------------
def _plot_features(df : pnd.DataFrame, test_name : str):
    out_dir = f'{Data.out_dir}/{test_name}'
    os.makedirs(out_dir, exist_ok=True)

    for feat, name in Data.d_feat.items() :
        df[feat].plot.hist(bins=100)

        plt.xlabel(name)
        plt.savefig(f'{out_dir}/{feat}.png')
        plt.close()
# ---------------------------------------------
def test_nobias(_dask_client : Client):
    '''
    Tests that:

    - The features can be retrieved
    - The bias is zero, i.e. mu=1
    '''
    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='fake_data', bias=1.0, kind='flat')

    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()
    df  = ddf.head(100)
    df['mu'] = df['mu'] / 1000.
    _plot_df(df=df, test_name='nobias', corr=None)

    arr_mu = df['mu'].to_numpy()

    assert numpy.allclose(arr_mu, 1, rtol=1e-5)
    assert set(df.columns) == Data.columns

    _plot_df(df=df, test_name='nobias', corr= None)
# ---------------------------------------------
@pytest.mark.parametrize('bias', [0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3])
def test_flat_bias(bias : float, _dask_client : Client):
    '''
    Tests that:

    - The features can be retrieved
    - The bias is the number that was injected
    '''
    corr= 1.0 / bias
    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='fake_data',bias=bias, kind='flat')

    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()
    df  = ddf.head(100)
    df['mu'] = df['mu'] / 1000.

    name = f'flat_bias/{bias}'
    _plot_df(df=df, test_name=name, corr=corr)

    arr_mu = df['mu'].to_numpy()

    assert numpy.allclose(arr_mu, corr, rtol=1e-5)
    assert set(df.columns) == Data.columns

    _plot_df(df=df, test_name=name, corr= None)
# ---------------------------------------------
def test_row_bias(_dask_client : Client):
    '''
    Tests that:

    - The features can be retrieved
    - The bias is the number that was injected
    '''
    bias = 1.0

    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='fake_data',bias=bias, kind='row')

    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()
    df  = ddf.head(100)
    df['mu'] = df['mu'] / 1000.

    name = f'row_bias/{100 * bias:.0f}'
    _plot_df(df=df, test_name=name, corr= None)

    assert set(df.columns) == Data.columns
# ---------------------------------------------
@pytest.mark.parametrize('bias', [0.9, 1.0, 1.2])
def test_features_target(_dask_client : Client, bias : float):
    '''
    Preprocesses a Dask dataframe and provides the tensor with the
    features
    '''
    log.info('Running test_features_target')

    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='fake_data', bias=bias, kind='flat')

    pre = PreProcessor(ddf=ddf, cfg=cfg)
    with pre.max_entries(100):
        fet = pre.features
        tgt = pre.targets / 1000

    nrows, ncols = fet.shape
    nsample, _   = tgt.shape

    assert ncols == 6
    assert nrows >  0
    assert nrows == nsample

    corr = 1.0 / bias
    assert torch.allclose(tgt, torch.tensor(corr), rtol=1e-5)
# ---------------------------------------------
def test_real_bias(_dask_client : Client):
    '''
    Run over real data
    '''
    cfg = cut.load_cfg(name='tests/preprocessor/simple')
    ddf = cut.get_ddf(name='real_data',bias=None, kind='flat')
    pre = PreProcessor(ddf=ddf, cfg=cfg)
    ddf = pre.get_data()

    df       = ddf.compute()
    df['mu'] = df['mu'] / 1000.
    _plot_df(df=df, test_name='real_bias', corr=None)
# ---------------------------------------------



'''
This module is used by pytest to _inject_ fixtures in the tests
'''
import os
import logging
import pytest
import mplhep
import matplotlib
import matplotlib.pyplot as plt

from dask.distributed      import Client
from dmu.logging.log_store import LogStore

# ---------------------------------------
def pytest_configure(config : pytest.Config) -> None:
    '''
    Runs before all tests, needed to do global initialization. e.g. logging level setting
    '''
    _ = config

    # Line below is needed to avoid core dump with Dask
    matplotlib.use('Agg')
    os.environ['ANADIR'] = '/tmp/tests/ecal_calibration'
    plt.style.use(mplhep.style.LHCb2)

    logging.getLogger("PIL.PngImagePlugin").setLevel(logging.WARNING)
    logging.getLogger("matplotlib.font_manager").setLevel(logging.WARNING)

    LogStore.set_level('ecal_calibration:regressor', 10)
# ---------------------------------------
@pytest.fixture(scope='session')
def _dask_client():
    '''
    Create a dask client to:

    - Turn off multithreading. Due to lack of thread safety in vector
    - Turn on multiprocessing
    '''
    client = Client(n_workers=12, threads_per_worker=1, processes=True)

    yield client
    client.close()
# ---------------------------------------



'''
Module with functions meant to test the row,col <-> x,y translation
'''
import os

import pytest
import pandas             as pnd
import matplotlib.pyplot  as plt

from matplotlib            import patches
from dmu.logging.log_store import LogStore
from ecal_calibration      import calo_translator as ctran

log=LogStore.add_logger('ecal_calibration:test_calo_translator')

subdetectors = [
            'Inner',
            'Middle',
            'Outer']
# --------------------------------
class Data:
    '''
    Data class
    '''
    out_dir = '/tmp/tests/rx_data/calo_translator'
# --------------------------------
def _plot_boundaries(ax):
    x1 = -2_000
    y1 = -1_200

    x2 =  2_000
    y2 =  1_200

    height= y2 - y1
    width = x2 - x1

    rectangle = patches.Rectangle((min(x1, x2), min(y1, y2)), width, height, edgecolor='red', facecolor='none', linewidth=2)
    ax.add_patch(rectangle)

    x1 = -1_000
    y1 =   -680

    x2 =  1_000
    y2 =   +680

    height= y2 - y1
    width = x2 - x1

    rectangle = patches.Rectangle((min(x1, x2), min(y1, y2)), width, height, edgecolor='red', facecolor='none', linewidth=2)
    ax.add_patch(rectangle)
# --------------------------------
def _plot_translation(df : pnd.DataFrame, row : int, col : int, det : str, name : str):
    if len(df) != 1:
        print(df)
        raise ValueError(f'Dataframe does not have one and only one element for: {row}/{col}/{name}')

    os.makedirs(f'{Data.out_dir}/{name}', exist_ok=True)

    ax = df.plot.scatter(x='x', y='y', color='blue', title=f'Det={det}; Row={row}; Col={col}')
    ax.set_xlim(-6_000, +6_000)
    ax.set_ylim(-4_000, +4_000)

    _plot_boundaries(ax)

    fname = f'{det}_{row:03}_{col:03}.png'
    plt.savefig(f'{Data.out_dir}/{name}/{fname}')
    plt.close()
# --------------------------------
def _plot_region(df : pnd.DataFrame, det : str, name : str, color : str = 'blue', skip_save : bool = False, ax=None):
    ax=df.plot.scatter(x='x', y='y', color=color, s=0.1, title=det, ax=ax)
    ax.set_xlim(-4_000, +4_000)
    ax.set_ylim(-3_500, +3_500)

    if not skip_save:
        os.makedirs(f'{Data.out_dir}/{name}', exist_ok=True)
        plt.grid()
        plt.savefig(f'{Data.out_dir}/{name}/{det}.png')
        plt.close()

    return ax
# --------------------------------
@pytest.fixture(scope='session', autouse=True)
def _initialize():
    '''
    This function will execute before the tests
    '''
    os.makedirs(Data.out_dir, exist_ok=True)
    LogStore.set_level('rx_data:calo_translator', 10)
# --------------------------------
def test_load():
    '''
    Tests loading data to dataframe
    '''
    log.info('')
    df = ctran.get_data()
    print(df)
    print(df.dtypes)
# --------------------------------
def test_read_xy():
    '''
    Tests reading XY coordinates, given row and column
    '''
    row = 10
    col = 10

    df     = ctran.get_data()
    in_row = df.r == row
    in_col = df.c == col

    df     = df[in_row & in_col]

    print(df)
# --------------------------------
def test_plot_row_col():
    '''
    Plots row and column histograms
    '''
    df     = ctran.get_data()
    df.c.hist(bins=64, histtype='step', linestyle='-', label='column')
    df.r.hist(bins=64, histtype='step', linestyle=':', label='row')
    plt.legend()
    plt.savefig(f'{Data.out_dir}/row_col.png')
    plt.close()
# --------------------------------
def test_plot_xyz():
    '''
    Tests plotting X and Y coodinates
    '''
    df     = ctran.get_data()
    df.x.hist(bins=50, histtype='step', label='x')
    df.y.hist(bins=50, histtype='step', label='y')
    df.z.hist(bins=50, histtype='step', label='z')
    plt.legend()
    plt.savefig(f'{Data.out_dir}/xyz.png')
    plt.close()
# --------------------------------
@pytest.mark.parametrize('det', subdetectors)
def test_scan_dets(det : str):
    '''
    Tests translation from row and column to x and y
    '''
    df = ctran.from_id_to_xy(det=det)

    _plot_region(df, det, name='detector_scan')
# --------------------------------
def test_overlay_detectors():
    '''
    Tests translation from row and column to x and y
    '''

    ax = None
    for det, color in [('Inner', 'red'), ('Middle', 'green'), ('Outer', 'blue')]:
        df = ctran.from_id_to_xy(det=det)
        ax = _plot_region(df, det, name='detector_scan', color=color, skip_save = True, ax=ax)

    plt.grid()
    plt.savefig(f'{Data.out_dir}/overlay_detectors.png')
    plt.close()
# --------------------------------
def test_fully_defined():
    '''
    Tests translation from row and column to x and y
    '''
    x1, y1 = ctran.from_id_to_xy(row=14, col=35, det ='Inner')
    x2, y2 = ctran.from_id_to_xy(row=14, col=35, area=      2)

    assert x1 == x2
    assert y1 == y2
# --------------------------------



'''
Module containing the Corrector class
'''
import torch
import pandas as pnd

from vector                     import MomentumObject4D as v4d
from dmu.logging.log_store      import LogStore
from ecal_calibration.network   import Network
from ecal_calibration.regressor import Regressor

log=LogStore.add_logger('ecal_calibration:corrector')
# ---------------------------------------
class Corrector:
    '''
    Class intended to calibrate electrons
    '''
    # ---------------------------------------
    def __init__(self, cfg : dict):
        '''
        Parameters:

        cfg: Dictionary with configuration
        '''
        self._cfg = cfg
        self._net = self._get_network()

        self._electron_mass = 0.511
    # ---------------------------------------
    def _get_network(self) -> Network:
        model_dir = Regressor.get_out_dir(cfg=self._cfg)
        net       = Regressor.load(model_dir=model_dir)
        if net is None:
            raise FileNotFoundError(f'Model could not be found in: {model_dir}')

        return net
    # ---------------------------------------
    def _get_correction_from_nn(self, row : pnd.Series) -> float:
        features = row.to_numpy()
        features = torch.tensor(features, dtype=torch.float32)

        targets  = self._net(features)
        val      = targets.detach().numpy()

        return val
    # ---------------------------------------
    def run(
            self,
            electron : v4d,
            row      : pnd.Series,
            from_nn  : bool = True) -> v4d:
        '''
        Calibrates electron

        Parameters
        -----------------
        electron: Lorentz vector before calibration
        row     : Pandas series with the features needed for predicting correction
        from_nn : If true, it will evaluate an NN, otherwise it will read the correction from input series

        Returns
        -----------------
        Lorentz vector representing calibrated electron
        '''
        if from_nn:
            val = self._get_correction_from_nn(row=row)
        else:
            val = row['mu']

        # The target was scaled by 1000 for training
        # Scale back the correction
        val         = float(val) / 1000.

        log.debug(f'Original : {electron}')
        res = v4d(
                pt  = val * electron.pt,
                eta =       electron.eta,
                phi =       electron.phi,
                mass= self._electron_mass)

        log.debug(f'Corrected: {res}')

        return res
# ---------------------------------------



'''
Module with tests for utility functions
'''
import os
import numpy
import pytest
import pandas as pnd
import matplotlib.pyplot as plt

from ecal_calibration   import utilities as cut

pytestmark = pytest.mark.skip(
    reason='''
    This part of the code did not produce useful results
    will turn off the tests, due to time constraints
    ''')
# -----------------------------------------
class Data:
    '''
    Data class
    '''
    out_dir = '/tmp/tests/ecal_calibration/utilities'
# -----------------------------------------
@pytest.fixture(scope='session', autouse=True)
def _initialize():
    os.makedirs(Data.out_dir, exist_ok=True)
# -----------------------------------------
def _plot_distributions(df : pnd.DataFrame, test_name : str) -> None:
    _plot_vtx( test_name=test_name, df=df)
    _plot_brem(test_name=test_name, df=df)
    _plot_lept(test_name=test_name, df=df, name= 'BREMHYPOROW')
    _plot_lept(test_name=test_name, df=df, name= 'BREMHYPOCOL')
    _plot_lept(test_name=test_name, df=df, name='BREMHYPOAREA')
    _plot_lept(test_name=test_name, df=df, name=          'PT')

    _plot_evt(test_name=test_name, df=df, name='nPVs')
    _plot_evt(test_name=test_name, df=df, name='block')
# -----------------------------------------
def _plot_evt(
        df        : pnd.DataFrame,
        test_name : str,
        name      : str) -> None:
    out_dir = f'{Data.out_dir}/{test_name}'
    os.makedirs(out_dir, exist_ok=True)

    df[name].plot.hist(bins=30)

    plt.legend()
    plt.xlabel(name)
    plt.savefig(f'{out_dir}/{name}.png')
    plt.close()
# -----------------------------------------
def _plot_lept(
        df        : pnd.DataFrame,
        test_name : str,
        name      : str) -> None:

    out_dir = f'{Data.out_dir}/{test_name}'
    os.makedirs(out_dir, exist_ok=True)

    if name == 'PT':
        arr_l1 = df[f'L1_{name}'].to_numpy()
        arr_l2 = df[f'L2_{name}'].to_numpy()

        arr_l1_log = numpy.log(arr_l1)
        arr_l2_log = numpy.log(arr_l2)

        plt.hist(arr_l1_log, range=[-5, +4], label='$e^+$', bins=100, alpha=0.3)
        plt.hist(arr_l2_log, range=[-5, +4], label='$e^-$', bins=100, histtype='step')

        plt.xlabel(f'Log {name}')
    else:
        df[f'L1_{name}'].plot.hist(label='$e^+$', bins=100, alpha=0.3)
        df[f'L2_{name}'].plot.hist(label='$e^-$', bins=100, histtype='step')

        plt.xlabel(name)

    plt.legend()
    plt.savefig(f'{out_dir}/{name}.png')
    plt.close()
# -----------------------------------------
def _plot_brem(df : pnd.DataFrame, test_name : str) -> None:
    out_dir = f'{Data.out_dir}/{test_name}'
    os.makedirs(out_dir, exist_ok=True)

    df['nbrem'  ] = df['L1_brem'] + df['L2_brem']

    df['L1_brem'].plot.hist(label='$e^+$', bins=10, alpha=0.3)
    df['L2_brem'].plot.hist(label='$e^-$', bins=10, alpha=0.3)
    df['nbrem'  ].plot.hist(label='both' , bins=10, alpha=0.3)

    plt.legend()
    plt.xlabel('brem')
    plt.savefig(f'{out_dir}/brem.png')
    plt.close()
# -----------------------------------------
def _plot_vtx(df : pnd.DataFrame, test_name : str) -> None:
    out_dir = f'{Data.out_dir}/{test_name}'
    os.makedirs(out_dir, exist_ok=True)

    df['B_END_VX'].plot.hist(label='VX', bins=80, alpha=0.3, range=[-20, +20])
    df['B_END_VY'].plot.hist(label='VY', bins=80, alpha=0.3, range=[-20, +20])
    df['B_END_VY'].plot.hist(label='VY', bins=80, alpha=0.3, range=[-20, +20])

    plt.yscale('log')
    plt.xlabel('B vertex position')
    plt.savefig(f'{out_dir}/end_vtx.png')
    plt.close()
# -----------------------------------------
@pytest.mark.parametrize('bias', [0.8, 1.0, 1.2])
def test_get_ddf_flat_bias(bias : float):
    '''
    Tests getter of dask dataframe
    '''
    ddf = cut.get_ddf(name='fake_data', bias=bias, kind='flat')
    df  = ddf.compute()

    assert len(df) == 100_000

    bias_str = f'{100 * bias:.0f}'
    _plot_distributions(df=df, test_name=f'nobias/{bias_str}')
# -----------------------------------------
